/* Copyright 1998 Dean Gaudet, all rights reserved.  Contact
 * dgaudet@arctic.org for more information.
 */

#include "flow.h"

#include <signal.h>

/* utility function: log an error including the NSPR error code */
void log_prerror(const char *fmt, ...)
{
    PRInt32 etext_len;
    char buf[1024];
    va_list args;
    PRInt32 buflen;
    PRInt32 rc;

#if 0
    etext_len = PR_GetErrorTextLength();
    if (etext_len >= sizeof(buf)) {
	buflen = PR_snprintf(buf, sizeof(buf), "error text for error code %u too long", PR_GetError());
	if (buflen == -1) {
	    PR_Abort();
	}
    }
    else {
	PR_GetErrorText(buf);
	buflen = strlen(buf);
    }
#else
    buflen = PR_snprintf(buf, sizeof(buf), "error %d", PR_GetError());
#endif
    if (buflen + 2 < sizeof(buf) - 1) {
	buf[buflen++] = ':';
	buf[buflen++] = ' ';
    }
    va_start(args, fmt);
    rc = PR_vsnprintf(buf + buflen, sizeof(buf) - buflen, fmt, args);
    /* XXX: kind of messy if it overflows */
    if (rc > 0) {
	buflen += rc;
    }
    buf[buflen] = '\n';
    PR_Write(PR_STDOUT, buf, buflen+1);
}

static PRFileDesc *make_sock(void)
{
    PRFileDesc *s;
    PRSocketOptionData sod;
    PRNetAddr sa_server;

    if ((s = PR_NewTCPSocket()) == NULL) {
	log_prerror("PR_NewTCPSocket()");
	return NULL;
    }

#define sodme(opt, v)	\
    sod.option = opt;	\
    sod.value.v;	\
    if (PR_SetSocketOption(s, &sod) != PR_SUCCESS) {	\
	log_prerror("PR_SetSocketOption(" #opt ")");	\
	return NULL;	\
    }
    sodme(PR_SockOpt_Reuseaddr, reuse_addr = 1);
    sodme(PR_SockOpt_Keepalive, keep_alive = 1);
    sodme(PR_SockOpt_NoDelay, no_delay = 1);
    sodme(PR_SockOpt_Nonblocking, non_blocking = 1);
#undef sodme

    if (PR_InitializeNetAddr(PR_IpAddrAny, 8080, &sa_server) != PR_SUCCESS) {
	log_prerror("PR_InitializeNetAddr(PR_IpAddrAny, 8080, &sa_server)");
	return NULL;
    }
    if (PR_Bind(s, &sa_server) != PR_SUCCESS) {
	log_prerror("PR_Bind(s, &sa_server)");
	return NULL;
    }
    if (PR_Listen(s, 127) != PR_SUCCESS) {
	log_prerror("PR_Listen(s, 127)");
	return NULL;
    }
    return s;
}

static PRIntn PR_CALLBACK nspr_main(PRIntn argc, char *argv[])
{
    PRThread *t_process;
    PRFileDesc *listen_socket;

    listen_socket = make_sock();
    if (listen_socket == NULL) {
	return 1;
    }

    t_process = PR_CreateThread(PR_USER_THREAD, request_loop,
	listen_socket, PR_PRIORITY_NORMAL, PR_LOCAL_THREAD,
	PR_UNJOINABLE_THREAD, 0);
    if (t_process == NULL) {
	log_prerror("PR_CreateThread(request_loop)");
	return 1;
    }

    /* wait forever */
    for(;;) {
	PR_Sleep(PR_SecondsToInterval(999999));
    }
    return 0;
}

static void term(int sig)
{
    exit(0);
}

int main(int argc, char **argv)
{
    signal(SIGINT, term);
    return PR_Initialize(nspr_main, argc, argv, 0);
}
