/* Copyright 1998 Dean Gaudet, all rights reserved.  Contact
 * dgaudet@arctic.org for more information.
 */

#include "flow.h"
#include "mqueue.h"

/* return non-zero when done sending... */
static int send_response(a_request *r)
{
    PRInt32 amt;
    int i;
    PRIOVec *v;

    i = r->cur_output_vec;
    v = r->output_vec;
    amt = PR_Writev(r->client, v + i, RESPONSE_VECMAX - i, PR_INTERVAL_NO_WAIT);
    if (amt < 0 && PR_GetError() != PR_WOULD_BLOCK_ERROR) {
	log_prerror("send_response: PR_Send returned -1");
	PR_Close(r->client);
	return 1;
    }
    if (amt <= 0) {
	return 0;
    }
    while (amt > 0) {
	if (amt < v[i].iov_len) {
	    v[i].iov_base = (char *) v[i].iov_base + amt;
	    v[i].iov_len -= amt;
	    r->cur_output_vec = i;
	    return 0;
	}
	amt -= v[i].iov_len;
	++i;
    }
    if (i == RESPONSE_VECMAX) {
	PR_Close(r->client);
	return 1;
    }
    r->cur_output_vec = i;
    return 0;
}


static const char month_snames[12][4] =
{
    "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};
static const char day_snames[7][4] =
{
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};


#define TOPSIZE	128

static PRInt32 setup_top(char *top_of_response)
{
    PRExplodedTime tms;

    /* XXX: this isn't exactly correct... or fast... it's just a hack
     * for the prototype.  In the real version we can have a thread
     * which maintains this info.
     */
    PR_ExplodeTime(PR_Now(), PR_GMTParameters, &tms);
    return PR_snprintf(top_of_response, TOPSIZE,
	    "HTTP/1.1 200 OK\r\n"
	    "Date: %s, %.2d %s %d %.2d:%.2d:%.2d GMT\r\n"
	    "Server: flow/0.1\r\n",
	    day_snames[tms.tm_wday],
	    tms.tm_mday, month_snames[tms.tm_month], tms.tm_year,
	    tms.tm_hour, tms.tm_min, tms.tm_sec);
}

void response_loop(void *dummy)
{
    a_message_queue *mq = dummy;
    a_request requests;
    PRInt32 n_requests;
    PRPollDesc *pds;
    PRInt32 n_pds;
    int i;
    a_request *r;
    a_request *nr;
    char top_of_response[TOPSIZE];
    int top_of_response_len;
    PRIntervalTime poll_interval;
    PRInt32 n_ready;

    poll_interval = PR_MillisecondsToInterval(100);

    n_requests = 0;
    PR_INIT_CLIST(&requests);
    n_pds = 100;	/* XXX: just an initial guess */
    pds = PR_Malloc(sizeof(*pds) * n_pds);
    if (pds == NULL) {
	PR_Abort();
    }

    for (;;) {
	if (n_requests == 0) {
	    r = mqueue_dequeue_en_masse(mq, 1);
	    top_of_response_len = setup_top(top_of_response);
	    goto process_new_requests;
	}

	/* fill in pds, we might need to grow it */
	if (n_requests > n_pds) {
	    PR_Free(pds);
	    n_pds = n_requests + 1;
	    pds = PR_Malloc(sizeof(*pds) * n_pds);
	    if (pds == NULL) {
		PR_Abort();
	    }
	}

	i = 0;
	r = PR_LIST_HEAD(&requests);
	while (i < n_requests) {
	    pds[i].fd = r->client;
	    pds[i].in_flags = PR_POLL_WRITE;
	    r = PR_NEXT_LINK(r);
	    ++i;
	}

	n_ready = PR_Poll(pds, n_requests, poll_interval);
	if (n_ready < 0) {
	    log_prerror("response_loop: PR_Poll");
	    return;
	}

	top_of_response_len = setup_top(top_of_response);

	i = 0;
	r = PR_LIST_HEAD(&requests);
	while (n_ready) {
	    nr = PR_NEXT_LINK(r);
	    if (pds[i].out_flags & (PR_POLL_EXCEPT|PR_POLL_ERR|PR_POLL_NVAL)) {
		/* XXX: do something more correct */
		log_prerror("response_loop: error on socket, discarding request");
		PR_REMOVE_LINK(r);
		PR_Free(r);
		--n_requests;
		--n_ready;
	    }
	    else if (pds[i].out_flags & PR_POLL_WRITE) {
		if (send_response(r)) {
		    PR_REMOVE_LINK(r);
		    PR_Free(r);
		    --n_requests;
		}
		--n_ready;
	    }
	    ++i;
	    r = nr;
	}

	r = mqueue_dequeue_en_masse(mq, 0);
process_new_requests:
	while (r) {
	    r->cur_output_vec = 0;
	    r->output_vec[0].iov_base = top_of_response;
	    r->output_vec[0].iov_len = top_of_response_len;
	    r->output_vec[1] = r->flow->headers_out;
	    r->output_vec[2] = r->flow->body;
	    nr = r->next;
	    if (send_response(r)) {
		PR_Free(r);
	    }
	    else {
		PR_APPEND_LINK(r, &requests);
		++n_requests;
	    }
	    r = nr;
	}
    }
}
