/* Copyright 1998 Dean Gaudet, all rights reserved.  Contact
 * dgaudet@arctic.org for more information.
 */

#ifndef FLOW_H_INCLUDED
#define FLOW_H_INCLUDED 1

#include <prtypes.h>
#include <prprf.h>
#include <prio.h>
#include <prthread.h>
#include <prinit.h>
#include <prerror.h>
#include <prclist.h>
#include <prnetdb.h>
#include <prmem.h>

/* flows are very basic in this proof-of-concept */
typedef struct a_flow a_flow;
struct a_flow {
    const char *url;
    size_t url_len;
    const char *filename;
    /* Any headers that are specific to this resource, such as
     * Last-Modified, Content-Length, and Content-Type.  Plus
     * the trailing \r\n to end the response headers.
     * Date and Server are automatically created on the fly.
     */
    PRIOVec headers_out;
    PRIOVec body;	/* the response body */
};

/* We don't have to handle arbitrary requests... anything that doesn't
 * fit in this size buffer will be punted to the full handling.
 */
#define REQUEST_MAXSIZE	2048

typedef struct {
    char *start;		/* points to first byte of substring */
    char *end;			/* points to first byte beyond last valid byte of substring */
} a_substring;

/* In the future we will be able to extend the headers supported
 * at run time... and in particular that means we'll be able to
 * support case-insensitivity... but maybe we don't want to support
 * that, maybe we'd rather just bounce the request.  A better hashing
 * function can be used later.
 */
typedef enum {
    HDR_accept = 0,
    HDR_accept_charset,
    HDR_accept_language,
    HDR_connection,
    HDR_host,
    HDR_user_agent,
    HDR_pragma,
    HDR_if_modified_since,
    HDR_bogus,
    HDR_MAX = HDR_bogus
} a_request_header_name;

typedef struct a_request a_request;
struct a_request {
    a_request *next;
    a_request *prev;
    PRFileDesc *client;
    PRNetAddr sa_client;

    a_substring url;
    a_substring hdrs[HDR_MAX];

    const a_flow *flow;

#define RESPONSE_VECMAX	3		/* used by response_loop */
    PRIOVec output_vec[RESPONSE_VECMAX];
    int cur_output_vec;

    char *end_request_buf;		/* pointer to first byte beyond last valid byte */

    unsigned char http_minor_version;	/* we handle HTTP/1.0 and HTTP/1.1 */
    char request_buf[REQUEST_MAXSIZE + 4];	/* + 4 for sentinels */
};

/* for printing i/o errors */
void log_prerror(const char *fmt, ...) __attribute__((format(printf,1,2)));

/* the main processing thread */
void request_loop(void *);
void response_loop(void *);

#endif
